# This code is Apache 2 licensed:
# https://www.apache.org/licenses/LICENSE-2.0

# This entire module was created by Simon Willison, and slightly modified by
# Dmitry Mazin. The blog post where Simon published the code is here:
# https://til.simonwillison.net/gpt3/chatgpt-api
import os
import openai


class ChatBot:
    def __init__(self, model, system="", is_verbose=False, api_key_path=None):
        self.system = system
        self.messages = []
        self.model: str = model
        self.is_verbose: bool = is_verbose

        # Precedence of ways to submit the API key:
        # 1. --api-key-path
        # 2. OPENAI_API_KEY environment variable
        # 3. ~/.openai-api-key
        if api_key_path:
            openai.api_key_path = api_key_path
            if self.is_verbose:
                print("Using API key from {}".format(api_key_path))
        elif os.getenv("OPENAI_API_KEY"):
            # Not actually setting the api key here, because OpenAI will itself
            # look for this env var
            if self.is_verbose:
                print("Using API key from OPENAI_API_KEY environment variable")
        elif api_key_path is None:
            default_key_path: str = os.path.join(
                os.path.expanduser("~"), ".openai-api-key"
            )
            if os.path.exists(default_key_path):
                with open(default_key_path) as f:
                    openai.api_key = f.read().strip()
                    if self.is_verbose:
                        print("Using API key from {}".format(default_key_path))
            else:
                raise ValueError(
                    "Please provide an API key path using --api-key-path or set the OPENAI_API_KEY environment variable."
                )

        if self.system:
            self.messages.append({"role": "system", "content": system})

    def __call__(self, message):
        self.messages.append({"role": "user", "content": message})
        result = self.execute()
        self.messages.append({"role": "assistant", "content": result})
        return result

    def execute(self):
        completion = openai.ChatCompletion.create(
            model=self.model, messages=self.messages
        )
        if self.is_verbose:
            # eg: {"completion_tokens": 86, "prompt_tokens": 26, "total_tokens": 112}
            print(completion.usage)
        return completion.choices[0].message.content



from setuptools import setup, find_packages

setup(
    name="git-lazy-commit",
    version='0.14.0',
    packages=find_packages(),
    entry_points={"console_scripts": ["git-lazy-commit = assistant:main"]},
    install_requires=["openai", "GitPython"],
)



import os
import argparse
import subprocess
import git
from .chatbot import ChatBot


class Assistant:
    def __init__(self, model="gpt-3.5-turbo", is_verbose=False, api_key_path=None):
        self.chatgpt = ChatBot(
            system="Please generate a commit message given the output of `git diff`. Please respond with a commit message without any additional text.",
            model=model,
            is_verbose=is_verbose,
            api_key_path=api_key_path,
        )

    def get_uncommitted_changes(self):
        staged_changes = subprocess.run(
            ["git", "diff", "--staged"], capture_output=True, text=True
        )
        uncommitted_changes = staged_changes.stdout.strip().split("\n")
        return uncommitted_changes

    def generate_commit_message(self, changes_summary):
        # Trim the changes summary because the most common model, gpt-3.5-turbo,
        # has a 4096 character limit
        return self.chatgpt(changes_summary[:4096])

    def generate_alternative_commit_message(self):
        return self.chatgpt(
            "Please come up with another commit message for the diff I sent earlier."
        )

    def commit_changes(self, repo, commit_message):
        repo.index.commit(commit_message)

    def get_user_approval(self, commit_msg):
        print(f"Generated commit message: {commit_msg}")
        user_input = (
            input("Do you approve this commit message? ((y)es/(n)o/(e)ditor): ")
            .strip()
            .lower()
        )

        if user_input in ["yes", "y"]:
            return True, commit_msg
        elif user_input in ["no", "n"]:
            return False, commit_msg
        elif user_input in ["editor", "e"]:
            return self.edit_commit_message(commit_msg)
        else:
            print(
                "Invalid input. Please enter 'yes (or y)', 'no (or n)', or 'editor (or e)'."
            )
            return self.get_user_approval(commit_msg)

    def edit_commit_message(self, commit_msg):
        with open("temp_commit_msg.txt", "w") as f:
            f.write(commit_msg)

        editor = os.environ.get("EDITOR", "vim")
        subprocess.run([editor, "temp_commit_msg.txt"])

        with open("temp_commit_msg.txt", "r") as f:
            edited_commit_msg = f.read().strip()

        os.remove("temp_commit_msg.txt")
        return self.get_user_approval(edited_commit_msg)


def main(args=None):
    if args is None:
        args = parse_arguments()

    assistant = Assistant(args.model, args.verbose, args.api_key_path)

    repo = git.Repo(os.getcwd())

    if args.add:
        repo.git.add(all=True)

    uncommitted_changes = assistant.get_uncommitted_changes()
    changes_summary = "\n".join(uncommitted_changes)

    if uncommitted_changes == [""]:
        print("There are no staged changes.")
        return

    generated_commit_message = assistant.generate_commit_message(changes_summary)

    while True:
        user_approved, commit_message = assistant.get_user_approval(
            generated_commit_message
        )
        if user_approved:
            break
        else:
            generated_commit_message = assistant.generate_alternative_commit_message()

    assistant.commit_changes(repo, commit_message)
    return "Changes committed."


def parse_arguments():
    parser = argparse.ArgumentParser(
        description="Generate commit messages based on staged changes."
    )
    parser.add_argument(
        "-m", "--model", default="gpt-3.5-turbo", help="OpenAI API model to use"
    )
    parser.add_argument(
        "--api-key-path",
        help="Tell me where you stored your OpenAI API key. If this isn't provided, I'll look for the OPENAPI_API_KEY env var, and failing that, ~/.openai_api_key.",
    )
    parser.add_argument(
        "--verbose", action="store_true", help="Print extra information"
    )
    parser.add_argument(
        "--add", action="store_true", help="Run `git add .` before generating the commit message"
    )
    args = parser.parse_args()
    return args



import unittest
from unittest.mock import MagicMock, patch, mock_open

from .context import assistant


class TestGitCommitAssistant(unittest.TestCase):
    def setUp(self):
        self.assistant = assistant.Assistant()

    def test_get_uncommitted_changes(self):
        with patch("subprocess.run") as run_mock:
            run_mock.return_value.stdout = "example change\n"
            changes = self.assistant.get_uncommitted_changes()
            self.assertEqual(changes, ["example change"])

    def test_generate_commit_message(self):
        changes_summary = "example change\n"
        with patch.object(
            self.assistant.chatgpt, "execute", return_value="Update example feature"
        ) as chatgpt_mock:
            commit_message = self.assistant.generate_commit_message(changes_summary)
            chatgpt_mock.assert_called_once()
            self.assertEqual(commit_message, "Update example feature")

    def test_commit_changes(self):
        repo_mock = MagicMock()
        commit_message = "Update example feature"
        self.assistant.commit_changes(repo_mock, commit_message)
        repo_mock.index.commit.assert_called_once_with(commit_message)

    def test_get_user_approval(self):
        commit_msg = "Update example feature"
        with patch("builtins.input", return_value="yes"):
            approval, message = self.assistant.get_user_approval(commit_msg)
            self.assertTrue(approval)
            self.assertEqual(message, commit_msg)

        with patch("builtins.input", return_value="no"):
            approval, message = self.assistant.get_user_approval(commit_msg)
            self.assertFalse(approval)
            self.assertEqual(message, commit_msg)

        with patch("builtins.input", side_effect=["invalid", "yes"]):
            approval, message = self.assistant.get_user_approval(commit_msg)
            self.assertTrue(approval)
            self.assertEqual(message, commit_msg)

        with patch.object(
            self.assistant,
            "edit_commit_message",
            return_value=(True, "Edited commit message"),
        ) as edit_mock:
            with patch("builtins.input", return_value="e"):
                approval, message = self.assistant.get_user_approval(
                    "Update example feature"
                )
                edit_mock.assert_called_once()
                self.assertTrue(approval)
                self.assertEqual(message, "Edited commit message")

    def test_generate_alternative_commit_message(self):
        with patch.object(
            self.assistant.chatgpt, "execute", return_value="Another commit message"
        ) as chatgpt_mock:
            alternative_commit_message = (
                self.assistant.generate_alternative_commit_message()
            )
            chatgpt_mock.assert_called_once()
            self.assertEqual(alternative_commit_message, "Another commit message")


if __name__ == "__main__":
    unittest.main()
